from __future__ import annotations

from typing import Optional

import numpy as np
import pandas as pd

from seeq.sdk import *
from seeq.spy import _common
from seeq.spy import _login
from seeq.spy import _metadata
from seeq.spy._errors import *
from seeq.spy._redaction import safely
from seeq.spy._session import Session
from seeq.spy._status import Status
from seeq.spy.workbooks import _folder
from seeq.spy.workbooks import _pull
from seeq.spy.workbooks import _user
from seeq.spy.workbooks._item import Item
from seeq.spy.workbooks._workbook import Workbook


def push(workbooks, *, path=None, owner=None, label=None, datasource=None, datasource_map_folder=None,
         use_full_path=False, access_control=None, override_max_interp=False,
         include_inventory=True, scope_globals_to_workbook=False, refresh=True,
         errors='raise', quiet=False, status: Optional[Status] = None, session: Optional[Session] = None):
    """
    Pushes workbooks into Seeq using a list of Workbook object definitions.

    Parameters
    ----------
    workbooks : {Workbook, list[Workbook]}
        A Workbook object or list of Workbook objects to be pushed into Seeq.

    path : str, default None
        A '>>'-delimited folder path to create to contain the workbooks. Note
        that a further subfolder hierarchy will be created to preserve the
        relative paths that the folders were in when they were searched for
        and pulled. If you specify None, then the workbook will stay where
        it is (if it has already been pushed once). If you specify
        spy.workbooks.MY_FOLDER, it will be moved to the user's home folder.
        If you specify a folder ID directly, it will be pushed to that folder.

    owner : str, default None
        Determines the ownership of pushed workbooks and folders.

        By default, the current owner will be preserved. If the content doesn't
        exist yet, the logged-in user will be the owner.

        All other options require that the logged-in user is an admin:

        If spy.workbooks.ORIGINAL_OWNER, ownership is assigned according to the
        original owner of the pulled content.

        If spy.workbooks.FORCE_ME_AS_OWNER, existing content will be
        reassigned to the logged-in user.

        If a username or a user's Seeq ID is supplied, that user will be
        assigned as owner.

        You may need to supply an appropriate datasource map if the usernames
        are different between the original and the target servers.

    label : str
        A user-defined label that differentiates this push operation from
        others. By default, the label will be the logged-in user's username
        OR the username from the 'owner' argument so that push activity will
        generally be isolated by user. But you can override this with a label
        of your choosing.

    datasource : str, optional
        The name of the datasource within which to contain all the
        pushed items. By default, all pushed items will be contained in a "Seeq
        Data Lab" datasource. Do not create new datasources unless you really
        want to and you have permission from your administrator.

    datasource_map_folder : str, default None
        A folder containing Datasource_Map_Xxxx_Yyyy_Zzzz.json files that can
        provides a means to map stored items (i.e., those originating from
        external datasources like OSIsoft PI) from one server to another or
        from one datasource to another (i.e., for workbook swapping). A default
        set of datasource map files is created during a pull/save sequence, and
        you can copy these default files to a folder, alter them, and then
        specify the folder as this argument.

    use_full_path : bool, default False
        If True, the original full path for an item is reconstructed, as
        opposed to the path that is relative to the Path property supplied to
        the spy.workbooks.search() call that originally helped create these
        workbook definitions. Note that this full path will still be inside
        the folder specified by the 'path' argument, if supplied.

    access_control : str, default None
        Specifies how Access Control Lists should be treated, via the
        following keywords: add/replace,loose/strict

        - If None, then no access control entries are pushed.
        - If 'add', then existing access control entries will not be disturbed
          but new entries will be added.
        - If 'replace', then existing access control entries will be removed
          and replaced with the entries from workbook definitions.
        - If 'loose', then any unmapped users/groups from the workbook
          definitions will be silently ignored.
        - If 'strict', then any unmapped users/groups will result in errors.

        Example: access_control='replace,loose'

    override_max_interp : bool, default False
        If True, then the Maximum Interpolation overrides from the source
        system will be written to the destination system.

    include_inventory : bool, default True
        If True, then all calculated items that are scoped to the workbook
        will be pushed as well.

    scope_globals_to_workbook : bool, default False
        If include_inventory=True and scope_globals_to_workbook=True, then
        all globally-scoped (aka "Available outside this analysis") items
        will be scoped to the workbook. False if you want to actually push
        these items to the global scope.

    refresh : bool, default True
        If True, then the Workbook objects that were supplied as input will
        be updated to be "fresh" from the server after the push. All
        identifiers will reflect the actual pushed IDs. Since refreshing
        takes time, you can set this to False if you don't plan to make
        further modifications to the Workbook objects or use the new IDs.

    errors : {'raise', 'catalog'}, default 'raise'
        If 'raise', any errors encountered will cause an exception. If
        'catalog', errors will be added to a 'Result' column in the status.df
        DataFrame (errors='catalog' must be combined with
        status=<Status object>).

    quiet : bool
        If True, suppresses progress output. Note that when status is
        provided, the quiet setting of the Status object that is passed
        in takes precedent.

    status : spy.Status, optional
        If specified, the supplied Status object will be updated as the command
        progresses. It gets filled in with the same information you would see
        in Jupyter in the blue/green/red table below your code while the
        command is executed. The table itself is accessible as a DataFrame via
        the status.df property.

    session : spy.Session, optional
        If supplied, the Session object (and its Options) will be used to
        store the login session state. This is useful to log in to different
        Seeq servers at the same time or with different credentials.

    """
    _common.validate_argument_types([
        (workbooks, 'workbooks', (Workbook, list)),
        (path, 'path', str),
        (owner, 'owner', str),
        (label, 'label', str),
        (datasource, 'datasource', str),
        (datasource_map_folder, 'datasource_map_folder', str),
        (use_full_path, 'use_full_path', bool),
        (access_control, 'access_control', str),
        (override_max_interp, 'override_max_interp', bool),
        (include_inventory, 'include_inventory', bool),
        (scope_globals_to_workbook, 'scope_globals_to_workbook', bool),
        (refresh, 'refresh', bool),
        (errors, 'errors', str),
        (quiet, 'quiet', bool),
        (status, 'status', Status),
        (session, 'session', Session)
    ])

    status = Status.validate(status, quiet)
    session = Session.validate(session)
    _login.validate_login(session, status)

    _common.validate_errors_arg(errors)

    owner_identity = None
    if owner not in [None, _user.ORIGINAL_OWNER, _user.FORCE_ME_AS_OWNER]:
        owner_identity = _login.find_user(session, owner)
        owner = owner_identity.id

    status.update('Pushing workbooks', Status.RUNNING)

    item_map = dict()

    if not isinstance(workbooks, list):
        workbooks = [workbooks]

    # Make sure the datasource exists
    datasource_output = _metadata.create_datasource(session, datasource)

    # Sort such that Analyses are pushed before Topics, since the latter usually depends on the former
    remaining_workbooks = list()
    sorted_workbooks = sorted(list(workbooks), key=lambda w: w['Workbook Type'])
    status.df = pd.DataFrame(columns=['ID', 'Name', 'Type', 'Workbook Type', 'Count', 'Time', 'Result'])
    for index in range(len(sorted_workbooks)):  # type: int
        workbook = sorted_workbooks[index]
        remaining_workbooks.append((index, workbook))
        status.df.at[index, 'ID'] = workbook.id if workbook.id else np.nan
        status.df.at[index, 'Name'] = workbook.name
        status.df.at[index, 'Type'] = workbook.type
        status.df.at[index, 'Workbook Type'] = workbook['Workbook Type']
        status.df.at[index, 'Count'] = 0
        status.df.at[index, 'Time'] = 0
        status.df.at[index, 'Result'] = 'Queued'

    datasource_maps = None
    if datasource_map_folder:
        datasource_maps = Workbook.load_datasource_maps(datasource_map_folder)
        for datasource_map in datasource_maps:
            # Specifying Override = True causes the StoredItem push code to try to look up an item based on the map
            # instead of directly using the item's ID. This allows for pulling a workbook and then pushing it with a
            # datasource map to swap the items within it.
            datasource_map['Override'] = True

    folder_id = _create_folder_path_if_necessary(session, path, errors, status)
    at_least_one_thing_pushed = False

    while len(remaining_workbooks) > 0:
        at_least_one_thing_pushed_this_iteration = False

        dependencies_not_found = list()
        for index, workbook in remaining_workbooks.copy():  # type: (int, Workbook)
            if not isinstance(workbook, Workbook):
                raise SPyRuntimeError('"workbooks" argument contains a non Workbook item: %s' % workbook)

            try:
                if label is None:
                    # If a label is not supplied, check to see if we should be automatically isolating by user.
                    isolate_by_user = False
                    if _common.get(workbook.definition, 'Isolate By User', default=False):
                        # Workbooks can be marked as 'Isolate By User' so they're not stepping on each other when
                        # they do spy.workbook.push(). All of the Example Exports are marked with "Isolate By User"
                        # equal to True.
                        isolate_by_user = True

                    if workbook.provenance == Item.CONSTRUCTOR:
                        # If a Workbook is constructed (and not persisted -- i.e., not pulled/loaded) then the best
                        # policy is to isolate such workbooks so that their Data IDs can't collide in the event that
                        # two users happen to name their workbooks the same.
                        isolate_by_user = True

                    if isolate_by_user:
                        label = owner_identity.username if owner_identity is not None else session.user.username

                status.reset_timer()

                status.current_df_index = index
                status.put('Count', 0)
                status.put('Time', 0)
                status.put('Result', 'Pushing')

                if datasource_maps is not None:
                    workbook.datasource_maps = datasource_maps

                workbook_folder_id = workbook.push_containing_folders(session, item_map, datasource_output,
                                                                      use_full_path, folder_id, owner, label,
                                                                      access_control, errors, status)

                try:
                    workbook.push(owner=owner, folder_id=workbook_folder_id, item_map=item_map, label=label,
                                  datasource=datasource, access_control=access_control,
                                  override_max_interp=override_max_interp, include_inventory=include_inventory,
                                  scope_globals_to_workbook=scope_globals_to_workbook, errors=errors, status=status,
                                  session=session)

                    at_least_one_thing_pushed = True
                    at_least_one_thing_pushed_this_iteration = True
                    remaining_workbooks.remove((index, workbook))
                    status.put('Result', 'Success')

                    if len(workbook.item_push_errors) > 0:
                        if errors == 'catalog':
                            status.warn(workbook.item_push_errors_str)
                        else:
                            raise SPyRuntimeError(workbook.item_push_errors_str)

                except SPyDependencyNotFound as e:
                    status.put('Count', 0)
                    status.put('Time', 0)
                    status.put('Result', f'Need dependency: {str(e)}')

                    dependencies_not_found.append(str(e))

            except BaseException as e:
                _common.raise_or_catalog(errors, status, e=e)

        if not at_least_one_thing_pushed_this_iteration:
            if errors == 'raise':
                raise SPyRuntimeError('Could not find the following dependencies:\n%s\n'
                                      'Therefore, could not import the following workbooks:\n%s\n' %
                                      ('\n'.join(dependencies_not_found),
                                       '\n'.join([str(workbook) for workbook in workbooks])))

            break

    for index in range(len(workbooks)):
        workbook = workbooks[index]
        status.update('[%s/%s] Fixing up workbook-to-workbook links: %s' %
                      (index + 1, len(workbooks), str(workbook)), Status.RUNNING)
        # If the workbook is a topic that has unresolved dependencies, the push will fail. We should have already
        # caught this when pushing the first time, so we can just continue here
        try:
            workbook.push_fixed_up_workbook_links(session, item_map, label, datasource_output, errors, status)
        except SPyDependencyNotFound:
            continue

    if refresh and at_least_one_thing_pushed:
        pull_df = status.df.copy()
        pull_df.drop(columns=['ID'], inplace=True)
        pull_df.rename(columns={'Pushed Workbook ID': 'ID'}, inplace=True)
        new_workbooks = _pull.pull(pull_df, include_inventory=include_inventory,
                                   errors=errors, status=status.create_inner('Refresh Workbook'), session=session)
        for workbook in workbooks:
            new_workbook = [w for w in new_workbooks if w.id == item_map[workbook.id]][0]
            workbook.refresh_from(new_workbook, item_map)
            if folder_id is not None and folder_id != _common.PATH_ROOT:
                workbook['Search Folder ID'] = folder_id

    unique_results = status.df['Result'].drop_duplicates()
    if len(unique_results) > 1 or (len(unique_results) == 1 and unique_results.iloc[0] != 'Success'):
        status.update('Errors encountered, look at Result column in returned DataFrame', Status.FAILURE)
    else:
        status.update('Push successful', Status.SUCCESS)

    return status.df


def _create_folder_path_if_necessary(session: Session, path, errors: str, status: Status):
    if _common.is_guid(path):
        return path

    folders_api = FoldersApi(session.client)

    if path is None:
        return None

    path = path.strip()

    if not path:
        return None

    if path == _folder.MY_FOLDER:
        return folders_api.get_folder(folder_id='mine').id

    workbook_path = _common.path_string_to_list(path)

    parent_id = None
    folder_id = None
    folder_filter = 'owner'
    for i in range(0, len(workbook_path)):
        existing_content_id = None
        content_name = workbook_path[i]

        if content_name in [_folder.SHARED, _folder.ALL, _folder.USERS]:
            raise SPyRuntimeError(f'"path" argument cannot contain {content_name} folder')

        if content_name == _folder.CORPORATE:
            if not session.corporate_folder:
                raise SPyRuntimeError(f'Attempting to push to Corporate folder but user does not have access')

            parent_id = session.corporate_folder.id
            folder_id = session.corporate_folder.id
            folder_filter = 'corporate'
            continue

        if parent_id:
            folders = safely(lambda: folders_api.get_folders(filter=folder_filter,
                                                             folder_id=parent_id,
                                                             limit=10000),
                             action_description=f'get Folders using filter "{folder_filter}" within {parent_id}',
                             additional_errors=[400],
                             errors=errors, status=status)  # type: WorkbenchItemOutputListV1
        else:
            folders = folders_api.get_folders(filter=folder_filter,
                                              limit=10000)  # type: WorkbenchItemOutputListV1

        if folders is not None:
            for content in folders.content:  # type: WorkbenchSearchResultPreviewV1
                if content.type == 'Folder' and content_name == content.name:
                    existing_content_id = content.id
                    break

        if not existing_content_id:
            folder_input = FolderInputV1()
            folder_input.name = content_name
            folder_input.description = 'Created by Seeq Data Lab'
            folder_input.owner_id = session.user.id
            folder_input.parent_folder_id = parent_id
            folder_output = safely(lambda: folders_api.create_folder(body=folder_input),
                                   action_description=f'create Folder {folder_input.name}',
                                   errors=errors, status=status)  # type: FolderOutputV1
            if folder_output is not None:
                existing_content_id = folder_output.id

        parent_id = existing_content_id
        folder_id = existing_content_id

    return folder_id



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 57.1.1-SNAPSHOT
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

from __future__ import absolute_import

import sys
import os
import re

# python 2 and python 3 compatibility library
from six import iteritems

from ..configuration import Configuration
from ..api_client import ApiClient
from ..models import *

class UsersApi(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    Ref: https://github.com/swagger-api/swagger-codegen
    """

    def __init__(self, api_client=None):
        config = Configuration()
        if api_client:
            self.api_client = api_client
        else:
            if not config.api_client:
                config.api_client = ApiClient()
            self.api_client = config.api_client

    def autocomplete_users_and_groups(self, **kwargs):
        """
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.autocomplete_users_and_groups(query=query_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param int offset: The pagination offset, the index of the first collection item that will be returned in this page of results
        :param int limit: The pagination limit, the total number of collection items that will be returned in this page of results
        :param str query: Part of user or group name to search for (required)
        :return: IdentityPreviewListV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: IdentityPreviewListV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.autocomplete_users_and_groups_with_http_info(**kwargs)
        else:
            (data) = self.autocomplete_users_and_groups_with_http_info(**kwargs)
            return data

    def autocomplete_users_and_groups_with_http_info(self, **kwargs):
        """
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.autocomplete_users_and_groups_with_http_info(query=query_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param int offset: The pagination offset, the index of the first collection item that will be returned in this page of results
        :param int limit: The pagination limit, the total number of collection items that will be returned in this page of results
        :param str query: Part of user or group name to search for (required)
        :return: IdentityPreviewListV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: IdentityPreviewListV1
        """

        all_params = ['offset', 'limit', 'query']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method autocomplete_users_and_groups" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'query' is set
        if ('query' not in params) or (params['query'] is None):
            raise ValueError("Missing the required parameter `query` when calling `autocomplete_users_and_groups`")


        collection_formats = {}

        path_params = {}
        if 'query' in params:
            path_params['query'] = params['query']

        query_params = []
        if 'offset' in params:
            query_params.append(('offset', params['offset']))
        if 'limit' in params:
            query_params.append(('limit', params['limit']))

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/users/autocomplete/{query}', 'GET',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='IdentityPreviewListV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def create_user(self, **kwargs):
        """
        Create a user
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.create_user(body=body_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param UserInputV1 body: User information (required)
        :return: UserOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UserOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.create_user_with_http_info(**kwargs)
        else:
            (data) = self.create_user_with_http_info(**kwargs)
            return data

    def create_user_with_http_info(self, **kwargs):
        """
        Create a user
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.create_user_with_http_info(body=body_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param UserInputV1 body: User information (required)
        :return: UserOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UserOutputV1
        """

        all_params = ['body']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method create_user" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'body' is set
        if ('body' not in params) or (params['body'] is None):
            raise ValueError("Missing the required parameter `body` when calling `create_user`")


        collection_formats = {}

        path_params = {}

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/users', 'POST',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='UserOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def delete_user(self, **kwargs):
        """
        Delete a user
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.delete_user(id=id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str id: ID of the user to delete (required)
        :param str new_owner_id: ID of the user to take over ownership of items. If not specified and the user to delete has owned items, a '400 Bad Request' will be returned.
        :param bool transfer_acl_and_group_membership: True if any ACL entries and group memberships for the user being deleted should be transferred to the new owner, false if ACL entries should be deleted and group memberships shouldn't be transferred. When migrating users to a different authentication method, this should be set to true. When actually deleting a user, it should usually be set to false. The default is 'false'.
        :return: StatusMessageBase
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: StatusMessageBase
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.delete_user_with_http_info(**kwargs)
        else:
            (data) = self.delete_user_with_http_info(**kwargs)
            return data

    def delete_user_with_http_info(self, **kwargs):
        """
        Delete a user
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.delete_user_with_http_info(id=id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str id: ID of the user to delete (required)
        :param str new_owner_id: ID of the user to take over ownership of items. If not specified and the user to delete has owned items, a '400 Bad Request' will be returned.
        :param bool transfer_acl_and_group_membership: True if any ACL entries and group memberships for the user being deleted should be transferred to the new owner, false if ACL entries should be deleted and group memberships shouldn't be transferred. When migrating users to a different authentication method, this should be set to true. When actually deleting a user, it should usually be set to false. The default is 'false'.
        :return: StatusMessageBase
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: StatusMessageBase
        """

        all_params = ['id', 'new_owner_id', 'transfer_acl_and_group_membership']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method delete_user" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'id' is set
        if ('id' not in params) or (params['id'] is None):
            raise ValueError("Missing the required parameter `id` when calling `delete_user`")


        collection_formats = {}

        path_params = {}
        if 'id' in params:
            path_params['id'] = params['id']

        query_params = []
        if 'new_owner_id' in params:
            query_params.append(('newOwnerId', params['new_owner_id']))
        if 'transfer_acl_and_group_membership' in params:
            query_params.append(('transferAclAndGroupMembership', params['transfer_acl_and_group_membership']))

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/users/{id}', 'DELETE',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='StatusMessageBase',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def get_me(self, **kwargs):
        """
        Get the user that belongs to this authentication session
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_me(callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :return: UserOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UserOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.get_me_with_http_info(**kwargs)
        else:
            (data) = self.get_me_with_http_info(**kwargs)
            return data

    def get_me_with_http_info(self, **kwargs):
        """
        Get the user that belongs to this authentication session
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_me_with_http_info(callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :return: UserOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UserOutputV1
        """

        all_params = []
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_me" % key
                )
            params[key] = val
        del params['kwargs']

        collection_formats = {}

        path_params = {}

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/users/me', 'GET',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='UserOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def get_user(self, **kwargs):
        """
        Get a user
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_user(id=id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str id: ID of the user to retrieve (required)
        :return: UserOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UserOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.get_user_with_http_info(**kwargs)
        else:
            (data) = self.get_user_with_http_info(**kwargs)
            return data

    def get_user_with_http_info(self, **kwargs):
        """
        Get a user
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_user_with_http_info(id=id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str id: ID of the user to retrieve (required)
        :return: UserOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UserOutputV1
        """

        all_params = ['id']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_user" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'id' is set
        if ('id' not in params) or (params['id'] is None):
            raise ValueError("Missing the required parameter `id` when calling `get_user`")


        collection_formats = {}

        path_params = {}
        if 'id' in params:
            path_params['id'] = params['id']

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/users/{id}', 'GET',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='UserOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def get_user_from_username(self, **kwargs):
        """
        Search for user from a username
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_user_from_username(auth_datasource_class=auth_datasource_class_value, auth_datasource_id=auth_datasource_id_value, username=username_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str auth_datasource_class: Auth datasource class of the user to retrieve (required)
        :param str auth_datasource_id: Auth datasource id of the user to retrieve (required)
        :param str username: Username of the user to retrieve (required)
        :return: UserOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UserOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.get_user_from_username_with_http_info(**kwargs)
        else:
            (data) = self.get_user_from_username_with_http_info(**kwargs)
            return data

    def get_user_from_username_with_http_info(self, **kwargs):
        """
        Search for user from a username
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_user_from_username_with_http_info(auth_datasource_class=auth_datasource_class_value, auth_datasource_id=auth_datasource_id_value, username=username_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str auth_datasource_class: Auth datasource class of the user to retrieve (required)
        :param str auth_datasource_id: Auth datasource id of the user to retrieve (required)
        :param str username: Username of the user to retrieve (required)
        :return: UserOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UserOutputV1
        """

        all_params = ['auth_datasource_class', 'auth_datasource_id', 'username']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_user_from_username" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'auth_datasource_class' is set
        if ('auth_datasource_class' not in params) or (params['auth_datasource_class'] is None):
            raise ValueError("Missing the required parameter `auth_datasource_class` when calling `get_user_from_username`")
        # verify the required parameter 'auth_datasource_id' is set
        if ('auth_datasource_id' not in params) or (params['auth_datasource_id'] is None):
            raise ValueError("Missing the required parameter `auth_datasource_id` when calling `get_user_from_username`")
        # verify the required parameter 'username' is set
        if ('username' not in params) or (params['username'] is None):
            raise ValueError("Missing the required parameter `username` when calling `get_user_from_username`")


        collection_formats = {}

        path_params = {}

        query_params = []
        if 'auth_datasource_class' in params:
            query_params.append(('authDatasourceClass', params['auth_datasource_class']))
        if 'auth_datasource_id' in params:
            query_params.append(('authDatasourceId', params['auth_datasource_id']))
        if 'username' in params:
            query_params.append(('username', params['username']))

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/users/search', 'GET',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='UserOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def get_users(self, **kwargs):
        """
        Get a collection of users
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_users(callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str username_search: Search text by which to filter users' names.
        :param str email_search: Search text by which to filter users' email addresses.
        :param str first_name_search: Search text by which to filter users' first names.
        :param str last_name_search: Search text by which to filter users' last names.
        :param str datasource_name_search: Search text by which to filter users' directories.
        :param bool is_admin: Whether to filter users to only users who are administrators.
        :param bool is_enabled: Whether to filter users to only users who are enabled.
        :param bool is_logged_in: Whether to filter users to only users who are logged in.
        :param bool is_created: Whether to filter to users that did not come pre-installed with Seeq.
        :param str sort_order: A field by which to order the users followed by a space and 'asc' or 'desc'. Field name can be one of: username, email, firstName, lastName, directory
        :param int offset: The pagination offset, the index of the first collection item that will be returned in this page of results
        :param int limit: The pagination limit, the total number of collection items that will be returned in this page of results
        :return: UserOutputListV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UserOutputListV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.get_users_with_http_info(**kwargs)
        else:
            (data) = self.get_users_with_http_info(**kwargs)
            return data

    def get_users_with_http_info(self, **kwargs):
        """
        Get a collection of users
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.get_users_with_http_info(callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param str username_search: Search text by which to filter users' names.
        :param str email_search: Search text by which to filter users' email addresses.
        :param str first_name_search: Search text by which to filter users' first names.
        :param str last_name_search: Search text by which to filter users' last names.
        :param str datasource_name_search: Search text by which to filter users' directories.
        :param bool is_admin: Whether to filter users to only users who are administrators.
        :param bool is_enabled: Whether to filter users to only users who are enabled.
        :param bool is_logged_in: Whether to filter users to only users who are logged in.
        :param bool is_created: Whether to filter to users that did not come pre-installed with Seeq.
        :param str sort_order: A field by which to order the users followed by a space and 'asc' or 'desc'. Field name can be one of: username, email, firstName, lastName, directory
        :param int offset: The pagination offset, the index of the first collection item that will be returned in this page of results
        :param int limit: The pagination limit, the total number of collection items that will be returned in this page of results
        :return: UserOutputListV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UserOutputListV1
        """

        all_params = ['username_search', 'email_search', 'first_name_search', 'last_name_search', 'datasource_name_search', 'is_admin', 'is_enabled', 'is_logged_in', 'is_created', 'sort_order', 'offset', 'limit']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_users" % key
                )
            params[key] = val
        del params['kwargs']


        collection_formats = {}

        path_params = {}

        query_params = []
        if 'username_search' in params:
            query_params.append(('usernameSearch', params['username_search']))
        if 'email_search' in params:
            query_params.append(('emailSearch', params['email_search']))
        if 'first_name_search' in params:
            query_params.append(('firstNameSearch', params['first_name_search']))
        if 'last_name_search' in params:
            query_params.append(('lastNameSearch', params['last_name_search']))
        if 'datasource_name_search' in params:
            query_params.append(('datasourceNameSearch', params['datasource_name_search']))
        if 'is_admin' in params:
            query_params.append(('isAdmin', params['is_admin']))
        if 'is_enabled' in params:
            query_params.append(('isEnabled', params['is_enabled']))
        if 'is_logged_in' in params:
            query_params.append(('isLoggedIn', params['is_logged_in']))
        if 'is_created' in params:
            query_params.append(('isCreated', params['is_created']))
        if 'sort_order' in params:
            query_params.append(('sortOrder', params['sort_order']))
        if 'offset' in params:
            query_params.append(('offset', params['offset']))
        if 'limit' in params:
            query_params.append(('limit', params['limit']))

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/users', 'GET',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='UserOutputListV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def update_password(self, **kwargs):
        """
        Update password for a user authenticating through Seeq
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.update_password(body=body_value, id=id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param UserPasswordInputV1 body: User information (required)
        :param str id: ID of the user to set password for (required)
        :return: UserOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UserOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.update_password_with_http_info(**kwargs)
        else:
            (data) = self.update_password_with_http_info(**kwargs)
            return data

    def update_password_with_http_info(self, **kwargs):
        """
        Update password for a user authenticating through Seeq
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.update_password_with_http_info(body=body_value, id=id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param UserPasswordInputV1 body: User information (required)
        :param str id: ID of the user to set password for (required)
        :return: UserOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UserOutputV1
        """

        all_params = ['body', 'id']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method update_password" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'body' is set
        if ('body' not in params) or (params['body'] is None):
            raise ValueError("Missing the required parameter `body` when calling `update_password`")
        # verify the required parameter 'id' is set
        if ('id' not in params) or (params['id'] is None):
            raise ValueError("Missing the required parameter `id` when calling `update_password`")


        collection_formats = {}

        path_params = {}
        if 'id' in params:
            path_params['id'] = params['id']

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/users/{id}/password', 'POST',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='UserOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)

    def update_user(self, **kwargs):
        """
        Update a user
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.update_user(body=body_value, id=id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param UserInputV1 body: User Update information (required)
        :param str id: ID of the user to update (required)
        :return: UserOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UserOutputV1
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('callback'):
            return self.update_user_with_http_info(**kwargs)
        else:
            (data) = self.update_user_with_http_info(**kwargs)
            return data

    def update_user_with_http_info(self, **kwargs):
        """
        Update a user
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please define a `callback` function
        to be invoked when receiving the response.
        >>> def callback_function(response):
        >>>     pprint(response)
        >>>
        >>> thread = api.update_user_with_http_info(body=body_value, id=id_value, callback=callback_function)

        :param callback function: The callback function
            for asynchronous request. (optional)
        :param UserInputV1 body: User Update information (required)
        :param str id: ID of the user to update (required)
        :return: UserOutputV1
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: UserOutputV1
        """

        all_params = ['body', 'id']
        all_params.append('callback')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method update_user" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'body' is set
        if ('body' not in params) or (params['body'] is None):
            raise ValueError("Missing the required parameter `body` when calling `update_user`")
        # verify the required parameter 'id' is set
        if ('id' not in params) or (params['id'] is None):
            raise ValueError("Missing the required parameter `id` when calling `update_user`")


        collection_formats = {}

        path_params = {}
        if 'id' in params:
            path_params['id'] = params['id']

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.\
            select_header_accept(['application/vnd.seeq.v1+json'])

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.\
            select_header_content_type(['application/vnd.seeq.v1+json', ])

        # Authentication setting
        auth_settings = ['api_key']

        return self.api_client.call_api('/users/{id}/update', 'POST',
                                        path_params,
                                        query_params,
                                        header_params,
                                        body=body_params,
                                        post_params=form_params,
                                        files=local_var_files,
                                        response_type='UserOutputV1',
                                        auth_settings=auth_settings,
                                        callback=params.get('callback'),
                                        _return_http_data_only=params.get('_return_http_data_only'),
                                        _preload_content=params.get('_preload_content', True),
                                        _request_timeout=params.get('_request_timeout'),
                                        collection_formats=collection_formats)



# coding: utf-8

"""
    Seeq REST API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 57.1.1-SNAPSHOT
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

from pprint import pformat
from six import iteritems
import re


class FormulaCompilerErrorOutputV1(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'column': 'int',
'error_category': 'str',
'error_type': 'str',
'function_id': 'str',
'item_id': 'str',
'line': 'int',
'message': 'str'    }

    attribute_map = {
        'column': 'column',
'error_category': 'errorCategory',
'error_type': 'errorType',
'function_id': 'functionId',
'item_id': 'itemId',
'line': 'line',
'message': 'message'    }

    def __init__(self, column=None, error_category=None, error_type=None, function_id=None, item_id=None, line=None, message=None):
        """
        FormulaCompilerErrorOutputV1 - a model defined in Swagger
        """

        self._column = None
        self._error_category = None
        self._error_type = None
        self._function_id = None
        self._item_id = None
        self._line = None
        self._message = None

        if column is not None:
          self.column = column
        if error_category is not None:
          self.error_category = error_category
        if error_type is not None:
          self.error_type = error_type
        if function_id is not None:
          self.function_id = function_id
        if item_id is not None:
          self.item_id = item_id
        if line is not None:
          self.line = line
        if message is not None:
          self.message = message

    @property
    def column(self):
        """
        Gets the column of this FormulaCompilerErrorOutputV1.
        The column of the formula that resulted in an error

        :return: The column of this FormulaCompilerErrorOutputV1.
        :rtype: int
        """
        return self._column

    @column.setter
    def column(self, column):
        """
        Sets the column of this FormulaCompilerErrorOutputV1.
        The column of the formula that resulted in an error

        :param column: The column of this FormulaCompilerErrorOutputV1.
        :type: int
        """

        self._column = column

    @property
    def error_category(self):
        """
        Gets the error_category of this FormulaCompilerErrorOutputV1.
        The category of the formula error, i.e. when it was encountered

        :return: The error_category of this FormulaCompilerErrorOutputV1.
        :rtype: str
        """
        return self._error_category

    @error_category.setter
    def error_category(self, error_category):
        """
        Sets the error_category of this FormulaCompilerErrorOutputV1.
        The category of the formula error, i.e. when it was encountered

        :param error_category: The error_category of this FormulaCompilerErrorOutputV1.
        :type: str
        """

        self._error_category = error_category

    @property
    def error_type(self):
        """
        Gets the error_type of this FormulaCompilerErrorOutputV1.
        The type of formula error that occurred

        :return: The error_type of this FormulaCompilerErrorOutputV1.
        :rtype: str
        """
        return self._error_type

    @error_type.setter
    def error_type(self, error_type):
        """
        Sets the error_type of this FormulaCompilerErrorOutputV1.
        The type of formula error that occurred

        :param error_type: The error_type of this FormulaCompilerErrorOutputV1.
        :type: str
        """

        self._error_type = error_type

    @property
    def function_id(self):
        """
        Gets the function_id of this FormulaCompilerErrorOutputV1.
        The function where the error occurred

        :return: The function_id of this FormulaCompilerErrorOutputV1.
        :rtype: str
        """
        return self._function_id

    @function_id.setter
    def function_id(self, function_id):
        """
        Sets the function_id of this FormulaCompilerErrorOutputV1.
        The function where the error occurred

        :param function_id: The function_id of this FormulaCompilerErrorOutputV1.
        :type: str
        """

        self._function_id = function_id

    @property
    def item_id(self):
        """
        Gets the item_id of this FormulaCompilerErrorOutputV1.
        The itemId that is the cause of the error

        :return: The item_id of this FormulaCompilerErrorOutputV1.
        :rtype: str
        """
        return self._item_id

    @item_id.setter
    def item_id(self, item_id):
        """
        Sets the item_id of this FormulaCompilerErrorOutputV1.
        The itemId that is the cause of the error

        :param item_id: The item_id of this FormulaCompilerErrorOutputV1.
        :type: str
        """

        self._item_id = item_id

    @property
    def line(self):
        """
        Gets the line of this FormulaCompilerErrorOutputV1.
        The line of the formula that resulted in an error

        :return: The line of this FormulaCompilerErrorOutputV1.
        :rtype: int
        """
        return self._line

    @line.setter
    def line(self, line):
        """
        Sets the line of this FormulaCompilerErrorOutputV1.
        The line of the formula that resulted in an error

        :param line: The line of this FormulaCompilerErrorOutputV1.
        :type: int
        """

        self._line = line

    @property
    def message(self):
        """
        Gets the message of this FormulaCompilerErrorOutputV1.
        An error message for the compiled formula

        :return: The message of this FormulaCompilerErrorOutputV1.
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """
        Sets the message of this FormulaCompilerErrorOutputV1.
        An error message for the compiled formula

        :param message: The message of this FormulaCompilerErrorOutputV1.
        :type: str
        """

        self._message = message

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, FormulaCompilerErrorOutputV1):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



from __future__ import annotations

import pandas as pd

from seeq.spy import _common, _search
from seeq.spy._errors import *
from seeq.spy._status import Status


def process_csv_data(data, status, workbook=_common.DEFAULT_WORKBOOK_PATH):
    """
    Processes a csv file into an appropriate dataframe for Tree constructor
    :param data: str
            Filename of a CSV
    :param status: spy.Status
            The status object to pass warnings to
    :param workbook: str, default 'Data Lab >> Data Lab Analysis'
            The path to a workbook (in the form of 'Folder >> Path >> Workbook Name')
            or an ID that all pushed items will be 'scoped to'. You can
            push to the Corporate folder by using the following pattern:
            '__Corporate__ >> Folder >> Path >> Workbook Name'. A Tree currently
            may not be globally scoped. These items will not be visible/searchable
            using the data panel in other workbooks.
    :return: pandas.Dataframe
    """

    status.update('Reading CSV data', Status.RUNNING)

    csv_data = None
    try:
        csv_data = pd.read_csv(data)
    except FileNotFoundError:
        message = (f"File {data} not found. Please ensure you have it in the correct working "
                   f"directory.")
        status.exception(SPyValueError(message), throw=True)
    except BaseException as e:
        message = f"Unexpected {e}, {type(e)}"
        status.exception(SPyValueError(message), throw=True)

    csv_columns = list(csv_data.columns)
    levels = [i for i in csv_columns if 'Level' in i]

    # validation
    if 'Name' not in csv_columns and 'ID' not in csv_columns:
        message = f"A 'Name' or 'ID' column is required"
        status.exception(SPyValueError(message), throw=True)

    join_column = get_complete_column_for_join(csv_data, csv_columns, status)

    if is_first_row_of_levels_columns_blank(csv_data, levels):
        message = f"All Level columns must have a value in the first row"
        status.exception(SPyValueError(message), throw=True)

    # forward fill Levels columns
    csv_data.loc[:, levels] = csv_data.loc[:, levels].ffill()

    if join_column == 'Name':
        csv_data = get_ids_by_name_from_user_input(csv_data, status, workbook=workbook)

    return csv_data


def is_first_row_of_levels_columns_blank(csv_data, levels) -> bool:
    return csv_data[levels].iloc[0].isnull().any()


def get_complete_column_for_join(csv_data, csv_columns, status):
    name_id = csv_data[[x for x in ['Name', 'ID'] if x in csv_columns]]
    join_cols = list(name_id.columns[~name_id.isnull().any()])
    join_col = None
    if len(join_cols) == 0:
        message = f"Either 'Name' or 'ID' column must be complete, without missing values."
        status.exception(SPyValueError(message), throw=True)
    elif len(join_cols) == 2:
        join_col = 'ID'
    else:
        join_col = join_cols[0]
    return join_col


def make_paths_from_levels(csv_data):
    """
    Gets the path from a series of levels columns, adds that to the
    dataframe, and removes the levels columns
    :param csv_data: pandas.Dataframe
    :return: None
    """

    def make_path(row, levels):
        path = " >> ".join([row[i] for i in levels if not is_nan_or_string_nan(row[i])])
        return path

    level_cols = [i for i in list(csv_data.columns) if 'Level' in i]
    if len(level_cols) == 0 and 'Path' in list(csv_data.columns):
        return

    elif len(level_cols) == 0:
        # there's no levels columns and no path
        raise SPyValueError(f"Levels columns or a path column must be provided")

    else:
        csv_data['Path'] = csv_data.apply(make_path, axis=1, levels=level_cols)
        csv_data.drop(columns=level_cols, inplace=True)


def is_nan_or_string_nan(item):
    if not pd.isnull(item):
        if isinstance(item, str):
            return item.lower() == "nan"
        return False
    return True


def get_ids_by_name_from_user_input(input_df, status, workbook=_common.DEFAULT_WORKBOOK_PATH):
    """
    Adds IDs from search to a dataframe, searches based on 'Name'
    :param input_df: pandas.Dataframe
    :param status: spy.Status
    :param workbook: str, default 'Data Lab >> Data Lab Analysis'
            The path to a workbook (in the form of 'Folder >> Path >> Workbook Name')
            or an ID that all pushed items will be 'scoped to'. You can
            push to the Corporate folder by using the following pattern:
            '__Corporate__ >> Folder >> Path >> Workbook Name'. A Tree currently
            may not be globally scoped. These items will not be visible/searchable
            using the data panel in other workbooks.
    :return: pandas.Dataframe
    """

    status.update('Searching for items from CSV', Status.RUNNING)

    _search_options = ['Name', 'Type']

    if 'ID' in input_df.columns:
        # rows with IDs, no search needed
        df_with_ids = input_df.dropna(subset=['ID'])

        # rows without IDs, to search by name
        df_no_ids = input_df[input_df['ID'].isnull()].copy()

    else:
        df_with_ids = pd.DataFrame()
        df_no_ids = input_df

    if 'ID' in df_no_ids.columns:
        df_no_ids.drop(columns=['ID'], inplace=True)

    # which of the options the csv has:
    search_cols = [x for x in _search_options if x in df_no_ids.columns]

    try:
        search_res = _search.search(
            df_no_ids[search_cols], order_by=["ID"], workbook=workbook, quiet=True
        )[['Name', 'ID']]
    except KeyError:
        status.warn(f"The following names did not return search results and were "
                    f"ignored: {list(df_no_ids['Name'])}")
        return df_with_ids

    # add warning for Names with multiple search results
    multiples = search_res.loc[search_res.duplicated(subset=['Name']), 'Name']
    if len(multiples) > 0:
        status.warn(f"The following names returned multiple search results, "
                    f"so the first result was used: {list(set(multiples))}")
        search_res = search_res.drop_duplicates(subset=['Name'], keep='first').reset_index(drop=True)

    # inner merge the search results with data from csv
    name_data = df_no_ids.merge(search_res, left_on='Name', right_on='Name')

    # add warning for Names without search results
    expected_names = set()
    found_names = set()
    found_names.update(name_data['Name'])
    expected_names.update(df_no_ids['Name'])
    missing_names = expected_names.difference(found_names)
    if len(missing_names) > 0:
        status.warn(f"The following names did not return search results and were "
                    f"ignored: {list(missing_names)}")

    all_data = pd.concat([name_data, df_with_ids], axis=0, ignore_index=True)

    return all_data
